
///////////////////////////////////////////////////////////////////////////////
//  This is SSI32 Random Number Generator at '08.09.30.                      //
//  SSI32 generates RNs nonrecursively(i.e. generates the k-th RN directly)  //
//  and requires a 64-bit C compiler [ bcc32(after 5.6) | icl | gcc | icc ]. //
///////////////////////////////////////////////////////////////////////////////
                                                                //('08.10.14)//

 /****************************************************************************/
 /*  Copyright_OML 2008 H. Yaguchi                                           */
 /*                                                                          */
 /*  The source code of this subroutine package can be obtained from ISMLIB  */
 /*       http://www.ism.ac.jp/computer_system/eng/ismlib/index.html         */
 /*  of the Institute of Statistical Mathematics without any charge.         */
 /*  On the conditions that terms of the OPEN MARKET LICENCE for software    */
 /*       http://www.ism.ac.jp/computer_system/eng/ismlib/license.html       */
 /*  are observed and that the copyright holder nor the Institute of         */
 /*  Statistical Mathematics take no responsibility on any result from the   */
 /*  use, this subroutine can be used or modified freely. On the condition   */
 /*  that this note is attached as it is, this subroutine can be             */
 /*  redistributed.  Appropriate reference must be made at times of          */
 /*  publishing results obtained using this subroutine.                      */
 /****************************************************************************/
       
//#define HowToUseSSI32    //Use this line if you want to see how to use SSI32.
//#define Use_icc          //Use this line if your compiler is icc.

#include <stdio.h>
#include <string.h>  //for memcpy, memset, memcmp

/////////////////////////////////////////////////////////
//   If your compiler is Microsoft's VC++, replace all //
//            long long int   below                    //
//         with                                        //
//            __int64.                                 //
/////////////////////////////////////////////////////////

/*============== Functions of SSI32K Random Number Generator ==============*/

unsigned int SSI32KBin32(void);   // generate a 32-bit RN by SSI32K
unsigned int SSI32KDec8(void);    // generate a decimal 8-digit RN by SSI32K
double       SSI32KU01(void);     // generate a unif RN in [0,1) by SSI32K
    // The period of SSI32K is  
    //    pq = 34359738337 x 34359738319 = 1.18 x 10^21.

/*----- SSI32Ksc : A short-cut version of SSI32K --------------------------*/
   // The time for gen. RNs is about 0.7 times of SSI32K.
   // (Use SSI32Ksc after some experience of SSI32K.)
unsigned int SSI32KscBin32(void); // generate a 32-bit  RN by SSI32Ksc
unsigned int SSI32KscDec8(void);  // generate a decimal 8-digit RN by SSI32Ksc
double       SSI32KscU01(void);   // generate a unif RN in [0,1) by SSI32Ksc

/*----- SSI32Ini : Initialization of SSI32 RNG ----------------------------*/
void SSI32IniK(unsigned long long int n);      
           // The initial position of generating RNs is set to n-th RN.
           // The n must be 0,1,2,... .
           // If SSI32Ini.. is not used, it is recognized that n = 0.
void SSI32IniKnm(unsigned long long int n, unsigned long long int m);
           // The initial position of generating RNs is set to (n*m)-th RN.
void SSI32IniKnml(unsigned long long int n, unsigned long long int m,
                  unsigned long long int l);   
           // The initial position of generating RNs is set to (n*m+l)-th RN.
void SSI32ChangeRandSeqK(unsigned long long int seqNo);
           // If you want to change the sequence of SSIK RNs, use this 
           // function.  This function changes the sequence of SSIK RNs to 
           // the [seqNo]-th seq. and sets parameters of SSIK so that SSIK
           // generates RNs starting from the 0-th RN.
           // It is checked that seq.s of SSIK RNs which have seqNo=0,1,...,
           // 1024*128 are different.

/*============ End of SSI32K Random Number Generator functions ============*/

/********** for SSI32 random number generation ********/
typedef struct
    {
    unsigned int L;  unsigned int H;    /* Intel's MPU(or equiv.) only */
    } uiHL;
typedef union
    {
    unsigned long long int I;
    uiHL HL;
    } uiLLHL;
typedef union
    {
    unsigned long long int I;
    uiHL HL;
    double Dbl;
    } uiLLHL_Dbl;
typedef struct
    {
    unsigned long long int p; unsigned long long int q;
    unsigned long long int r; unsigned long long int s;
    unsigned long long int rk; unsigned long long int sk;
    unsigned long long int XpBase; unsigned long long int XqBase;
    unsigned long long int w0; unsigned long long int v0;
    uiLLHL wkLL; uiLLHL vkLL;
    } SIPara64;
SIPara64 SIParaSSI32Korg =
    {
    0x7ffffffe1,       // p=34359738337
    0x7ffffffcf,       // q=34359738319
    0x39f750241,       // r=15560147521
    0x32f50fee9,       // s=13678739177
    0, 0,              // rk, sk
    0x88237449a,       // XpBase
    0xbdda73ad3,       // XqBase
    0x18237449a,       // w0
    0x1dda73ad3,       // v0
    0, 0               // wkLL, vkLL
    };
SIPara64 SIParaSSI32K =
    {
    0x7ffffffe1,       // p=34359738337
    0x7ffffffcf,       // q=34359738319
    0x39f750241,       // r=15560147521
    0x32f50fee9,       // s=13678739177
    0, 0,              // rk, sk
    0x88237449a,       // XpBase
    0xbdda73ad3,       // XqBase
    0x18237449a,       // w0
    0x1dda73ad3,       // v0
    0, 0               // wkLL, vkLL
    };

/********** for multiple-Integer computation used in SSI32Ini ***********/
#define  mIntBase       0x100000000
#define  mIntBaseM1     0x0ffffffff
#define  mIntBaseBits   32
#define  mIntIntSize    6
#define  mIntIndexMin   1
#define  mIntIndexMax   5
typedef struct  {unsigned int mIdgt[mIntIntSize];} mInt;
typedef struct  {unsigned int mIdgt[1+mIntIndexMax*2];} mIntD;

int mIntADD_HL(mIntD *UD, mIntD *VD, mIntD *UVD);
int mIntSUB_HL(mIntD *UD, mIntD *VD, mIntD *WD);
int mIntMUL_HL(mInt *U, mInt *V, mIntD *UVD);
int mIntDIV_HL(mIntD *UD, mInt *Vcl, mIntD *QD, mIntD *RD);
/*******************************************************/
unsigned int SSI32K(SIPara64 *SIParaL)
    {
    int    i;
    uiLLHL  WLL1, XLL1, WLL1_Wk;
    uiLLHL  WLL2, XLL2, WLL2_Wk;

    WLL1.I = SIParaL->w0;
    XLL1.I = SIParaL->XpBase;
    XLL1.I ^= (SIParaL->rk);

    WLL2.I = SIParaL->v0;
    XLL2.I = SIParaL->XqBase;
    XLL2.I ^= (SIParaL->sk);

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    SIParaSSI32K.wkLL.I = WLL1_Wk.I;
    SIParaSSI32K.vkLL.I = WLL2_Wk.I;
    WLL1_Wk.I -= WLL2_Wk.I;
    WLL1_Wk.I <<= 16;
    return(WLL1_Wk.HL.H);
    }
/**************/
unsigned int SSI32Ksc(SIPara64 *SIParaL)
    {
    int    i;
    uiLLHL  WLL1, XLL1, WLL1_Wk;
    uiLLHL  WLL2, XLL2, WLL2_Wk;

    WLL1.I = SIParaL->w0;
    XLL1.I = SIParaL->XpBase;
    XLL1.I ^= (SIParaL->rk);

    WLL2.I = SIParaL->v0;
    XLL2.I = SIParaL->XqBase;
    XLL2.I ^= (SIParaL->sk);

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    WLL1_Wk.I = WLL1.I * XLL1.I;
    WLL1.HL.L = WLL1_Wk.HL.H;
    WLL2_Wk.I = WLL2.I * XLL2.I;
    WLL2.HL.L = WLL2_Wk.HL.H;

    SIParaSSI32K.wkLL.I = WLL1_Wk.I;
    SIParaSSI32K.vkLL.I = WLL2_Wk.I;
    WLL1_Wk.I -= WLL2_Wk.I;
    WLL1_Wk.I <<= 16;
    return(WLL1_Wk.HL.H);
    }
/**************/
void SSI32plusRS(SIPara64 *SIParaL)
    {
    SIParaL->rk += SIParaL->r;
    if (SIParaL->rk >= SIParaL->p) {SIParaL->rk -= SIParaL->p;}
    SIParaL->sk += SIParaL->s;
    if (SIParaL->sk >= SIParaL->q) {SIParaL->sk -= SIParaL->q;}
    }
/**************/
void SSI32IniValidityCheck(mIntD *UD, mInt *Vcl, mIntD *QD, mIntD *RD, char ChkId)
    {
    mInt   U;
    mIntD  UVD, wkD;
    int    i, err=0;
    uiLLHL LLHLV, LLHLR;

    memcpy(&U, &QD->mIdgt[mIntIndexMax], sizeof(mInt));
    U.mIdgt[mIntIndexMin-1] = 0;
    mIntMUL_HL(&U, Vcl, &UVD);
    mIntADD_HL(&UVD, RD, &wkD);
    if (memcmp((char *)&wkD, (char *)UD, sizeof(mIntD)) != 0) {err=1;}

    LLHLV.HL.H = Vcl->mIdgt[mIntIndexMax-1];
    LLHLV.HL.L = Vcl->mIdgt[mIntIndexMax];
    LLHLR.HL.H = RD->mIdgt[mIntIndexMax*2-1];
    LLHLR.HL.L = RD->mIdgt[mIntIndexMax*2];
    if (LLHLV.I<=LLHLR.I) {err=2;}

    if (err != 0)
        {
        printf("Validity-Error detected by SSI32IniValidityCheck() [%ck, err#=%d]\n", ChkId, err);
        printf("\n[UD] ");
        for (i=0; i<=mIntIndexMax*2; i++) {printf("%08x ", UD->mIdgt[i]);}
        printf("\n[Vcl]");
        for (i=0; i<=mIntIndexMax; i++) {printf("%08x ", Vcl->mIdgt[i]);}
        printf("\n[QD] ");
        for (i=0; i<=mIntIndexMax*2; i++) {printf("%08x ", QD->mIdgt[i]);}
        printf("\n[RD] ");
        for (i=0; i<=mIntIndexMax*2; i++) {printf("%08x ", RD->mIdgt[i]);}
        printf("\n[UVD]");
        for (i=0; i<=mIntIndexMax*2; i++) {printf("%08x ", UVD.mIdgt[i]);}
        printf("\n[wkD]");
        for (i=0; i<=mIntIndexMax*2; i++) {printf("%08x ", wkD.mIdgt[i]);}
        printf("\n\nIf your compiler is icc, use the compiler option -O0 and try again.\n\n");
        }
    }
/**************/
void SSI32IniCom(SIPara64 *SIParaL, mInt *N)
     {
     mInt  U, V, Vcl;
     mIntD UVD, QD, RD;
     int   c;
     uiLLHL LLHLn, LLHLwk;

     memcpy(&U, N, sizeof(mInt));

     memset(&V, 0, sizeof(mInt));
     memset(&Vcl, 0, sizeof(mInt));
     memset(&UVD, 0, sizeof(mIntD));
     memset(&QD, 0, sizeof(mIntD));
     memset(&RD, 0, sizeof(mIntD));

     LLHLwk.I = SIParaL->r;  
     V.mIdgt[mIntIndexMax-1] = LLHLwk.HL.H;
     V.mIdgt[mIntIndexMax] = LLHLwk.HL.L;
     mIntMUL_HL(&U, &V, &UVD);
     LLHLwk.I = SIParaL->p;  
     Vcl.mIdgt[mIntIndexMax-1] = LLHLwk.HL.H;
     Vcl.mIdgt[mIntIndexMax] = LLHLwk.HL.L;
     mIntDIV_HL(&UVD, &Vcl, &QD, &RD);
     SSI32IniValidityCheck(&UVD, &Vcl, &QD, &RD, 'r');
     LLHLwk.HL.H = RD.mIdgt[mIntIndexMax*2-1];
     LLHLwk.HL.L = RD.mIdgt[mIntIndexMax*2];
     SIParaL->rk = LLHLwk.I;   // set rk

     LLHLwk.I = SIParaL->s;  
     V.mIdgt[mIntIndexMax-1] = LLHLwk.HL.H;
     V.mIdgt[mIntIndexMax] = LLHLwk.HL.L;
     mIntMUL_HL(&U, &V, &UVD);
     LLHLwk.I = SIParaL->q;  
     Vcl.mIdgt[mIntIndexMax-1] = LLHLwk.HL.H;
     Vcl.mIdgt[mIntIndexMax] = LLHLwk.HL.L;
     mIntDIV_HL(&UVD, &Vcl, &QD, &RD);
     SSI32IniValidityCheck(&UVD, &Vcl, &QD, &RD, 's');
     LLHLwk.HL.H = RD.mIdgt[mIntIndexMax*2-1];
     LLHLwk.HL.L = RD.mIdgt[mIntIndexMax*2];
     SIParaL->sk = LLHLwk.I;
     }

/**************/
void SSI32Ini(SIPara64 *SIParaL, unsigned long long int n)
     {
     uiLLHL LLHLn;
     mInt  U;

     memset(&U, 0, sizeof(mInt));
     LLHLn.I =n;  
     U.mIdgt[mIntIndexMax-1] = LLHLn.HL.H;
     U.mIdgt[mIntIndexMax] = LLHLn.HL.L;

     SSI32IniCom(SIParaL, &U);
     }

/**************/
void SSI32IniNM(SIPara64 *SIParaL, unsigned long long int n, 
                                   unsigned long long int m)
     {
     uiLLHL LLHLn;
     mInt  U, V;
     mIntD UVD;

     memset(&U, 0, sizeof(mInt));
     memset(&V, 0, sizeof(mInt));
     memset(&UVD, 0, sizeof(mIntD));

     LLHLn.I = n;  
     U.mIdgt[mIntIndexMax-1] = LLHLn.HL.H;
     U.mIdgt[mIntIndexMax] = LLHLn.HL.L;
     LLHLn.I = m;  
     V.mIdgt[mIntIndexMax-1] = LLHLn.HL.H;
     V.mIdgt[mIntIndexMax] = LLHLn.HL.L;
     mIntMUL_HL(&U, &V, &UVD);
     memcpy(&U, &UVD.mIdgt[mIntIndexMax], sizeof(mInt));

     SSI32IniCom(SIParaL, &U);
     }

/**************/
void SSI32IniNML(SIPara64 *SIParaL, unsigned long long int n, 
                  unsigned long long int m, unsigned long long int l)
     {
     uiLLHL LLHLn;
     mInt  U, V;
     mIntD UVD;
     mIntD UD, UVD2;

     memset(&U, 0, sizeof(mInt));
     memset(&V, 0, sizeof(mInt));
     memset(&UVD, 0, sizeof(mIntD));
     memset(&UD, 0, sizeof(mIntD));
     memset(&UVD2, 0, sizeof(mIntD));

     LLHLn.I = n;  
     U.mIdgt[mIntIndexMax-1] = LLHLn.HL.H;
     U.mIdgt[mIntIndexMax] = LLHLn.HL.L;
     LLHLn.I = m;  
     V.mIdgt[mIntIndexMax-1] = LLHLn.HL.H;
     V.mIdgt[mIntIndexMax] = LLHLn.HL.L;
     mIntMUL_HL(&U, &V, &UVD);
     LLHLn.I = l;  
     UD.mIdgt[mIntIndexMax*2-1] = LLHLn.HL.H;
     UD.mIdgt[mIntIndexMax*2] = LLHLn.HL.L;
     mIntADD_HL(&UD, &UVD, &UVD2);
     memcpy(&U, &UVD2.mIdgt[mIntIndexMax], sizeof(mInt));

     SSI32IniCom(SIParaL, &U);
     }

/**************/
void SSI32ChangeRandSeqK(unsigned long long int seqNo)
    {
    uiLLHL ranBin64P, ranBin64Q;
    memcpy(&SIParaSSI32K, &SIParaSSI32Korg, sizeof(SIParaSSI32K));
    if (seqNo==0) {return;}
    SSI32Ini(&SIParaSSI32K, (seqNo-1)*4);
    ranBin64P.HL.H=SSI32KBin32();
    ranBin64P.HL.L=SSI32KBin32();
    ranBin64P.I &= 0xfffffffff;
    ranBin64P.I |= 0x800000000;
    ranBin64Q.HL.H=SSI32KBin32();
    ranBin64Q.HL.L=SSI32KBin32();
    ranBin64Q.I &= 0xfffffffff;
    ranBin64Q.I |= 0x800000000;
    memcpy(&SIParaSSI32K, &SIParaSSI32Korg, sizeof(SIParaSSI32K));
    SIParaSSI32K.XpBase=ranBin64P.I;
    SIParaSSI32K.XqBase=ranBin64Q.I;
    }

/*===============================================================*/
int mIntMUL_HL(mInt *U, mInt *V, mIntD *UVD)
    {
    uiLLHL LLHLu, LLHLv, LLHLw;
    int i, j, iPj;
    unsigned long long int c, uv;
    uiLLHL LLMulAddWk;

    if (U->mIdgt[mIntIndexMin-1] != 0) {return(1);}
    if (V->mIdgt[mIntIndexMin-1] != 0) {return(2);}

    memset(UVD, 0, sizeof(mIntD));

    LLHLu.I = 0;
    LLHLv.I = 0;
    LLHLw.I = 0;

    for (j=mIntIndexMax; j>=mIntIndexMin; j--)       /* V */
        {
        if (V->mIdgt[j] == 0)
            {
            UVD->mIdgt[j] = 0;
            continue;
            }

        c=0;
        LLHLv.HL.L = V->mIdgt[j];
        for (i=mIntIndexMax; i>=mIntIndexMin; i--)   /* U */
            {
            iPj = i+j;
            LLHLu.HL.L = U->mIdgt[i];
            uv = LLHLv.I * LLHLu.I;   /* u*v */
            LLMulAddWk.I = (uv & mIntBaseM1);
            LLHLw.HL.L = UVD->mIdgt[iPj];
            LLMulAddWk.I += LLHLw.I;
            LLMulAddWk.I += c;
            UVD->mIdgt[iPj] = LLMulAddWk.HL.L;
            c = (uv >> mIntBaseBits);
            c += (LLMulAddWk.I >> mIntBaseBits);
            }
        UVD->mIdgt[j] = (unsigned int)c;
        }
    return(0);
    }
/*===============================================================*/
int mIntSUB_HL(mIntD *UD, mIntD *VD, mIntD *WD)
    {
    uiLLHL LLsubWk, LLHLu, LLHLv;
    int j;

    if (UD->mIdgt[mIntIndexMin-1] != 0) {return(1);}
    if (VD->mIdgt[mIntIndexMin-1] != 0) {return(2);}

    LLsubWk.I = 0;
    LLHLu.I = 0;
    LLHLv.I = 0;
    for (j=mIntIndexMax*2; j>=mIntIndexMin; j--)
        {
        LLHLu.HL.L = UD->mIdgt[j];
        LLsubWk.I += LLHLu.I;
        LLHLv.HL.L = VD->mIdgt[j];
        LLsubWk.I -= LLHLv.I;
        WD->mIdgt[j] = LLsubWk.HL.L;

        if (LLsubWk.HL.H != 0)
            { LLsubWk.I = -1; }
          else
            { LLsubWk.I = 0; }
        }

    WD->mIdgt[mIntIndexMin-1] = LLsubWk.HL.L;
    return(0);
    }
/*===============================================================*/
int mIntADD_HL(mIntD *UD, mIntD *VD, mIntD *WD)
    {
    uiLLHL LLaddWk, LLHLu, LLHLv;
    int j;

    if (UD->mIdgt[mIntIndexMin-1] != 0) {return(1);}
    if (VD->mIdgt[mIntIndexMin-1] != 0) {return(2);}

    LLaddWk.I = 0;
    LLHLu.I = 0;
    LLHLv.I = 0;
    for (j=mIntIndexMax*2; j>=mIntIndexMin; j--)
        {
        LLHLu.HL.L = UD->mIdgt[j];
        LLaddWk.I += LLHLu.I;
        LLHLv.HL.L = VD->mIdgt[j];
        LLaddWk.I += LLHLv.I;
        WD->mIdgt[j] = LLaddWk.HL.L;
        LLaddWk.I>>=mIntBaseBits;
        }

    WD->mIdgt[mIntIndexMin-1] = LLaddWk.HL.H;
    return(0);
    }
/*===============================================================*/
int mIntDIVs_HL(mInt *V, mIntD *QD, mIntD *RD)
    {
    unsigned long long int Dv, rbPr;
    int  j;

    // Situation  is   VleftPos=mIntIndexMax;

    Dv = V->mIdgt[mIntIndexMax];
    /* U is copied in R by mIntDIV() */

    QD->mIdgt[mIntIndexMin-1] = 0;
    for(j=mIntIndexMin-1; j<mIntIndexMax*2; j++)   /* R */
        {
        rbPr = (unsigned long long int)(RD->mIdgt[j]*mIntBase + RD->mIdgt[j+1]);
        RD->mIdgt[j] = 0;
        QD->mIdgt[j+1] = (unsigned int)(rbPr / Dv);
        RD->mIdgt[j+1] = (unsigned int)(rbPr % Dv);
        }

    return(0);
    }

/****************/
int mIntDIV_HL(mIntD *UD, mInt *Vcl, mIntD *QD, mIntD *RD)
    {
    uiLLHL IHLu, IHLv;
    mInt V;
    int i, j, iPj, D, M;
    int VleftPos;
    unsigned long long int vq, vqL, c;
    unsigned long long int rbPr, QHAT, RHAT, LHS, RHS;
    uiLLHL MulSubWk;

#ifdef Use_icc
    uiLLHL LLHLRh, LLHLL, LLHLR ;  //in order to avoid a bug of icc
    char   dmyMsg[200];
#endif

    if (UD->mIdgt[mIntIndexMin-1] != 0) {return(1);}
    if (Vcl->mIdgt[mIntIndexMin-1] != 0) {return(2);}

    VleftPos=1;
    for (j=mIntIndexMin; j<=mIntIndexMax; j++)    /* Vcl */
        {
        if(Vcl->mIdgt[j] == 0)
           { VleftPos ++; }
         else
           { break; }
        }

    if (VleftPos > mIntIndexMax) {return(3);}

    memcpy(&V, Vcl, sizeof(mInt));
    memcpy(RD, UD, sizeof(mIntD));
    memset(QD, 0, sizeof(mIntD));

    if (VleftPos == mIntIndexMax) {return(mIntDIVs_HL(Vcl, QD, RD));}
                                     /* U is copied in R */

    /* normalization */
    RD->mIdgt[mIntIndexMin-1] = 0;
    V.mIdgt[VleftPos-1] = 0;
    if((V.mIdgt[VleftPos] >> (mIntBaseBits/2)) == 0)
        {
        D = mIntBaseBits/2 ;
        for(i=VleftPos; i<mIntIndexMax; i++)
            {
            V.mIdgt[i] = (unsigned int)((V.mIdgt[i]<<D) | (V.mIdgt[i+1]>>D));
            }
        V.mIdgt[mIntIndexMax]<<=D;
        for(i=mIntIndexMin-1; i<mIntIndexMax*2; i++)
            {
            RD->mIdgt[i] = (unsigned int)((RD->mIdgt[i]<<D) | (RD->mIdgt[i+1]>>D));
            }
        RD->mIdgt[mIntIndexMax*2]<<=D;
        }
      else
        {
        D = 0;
        }
    
    M = VleftPos - mIntIndexMin;  /* M >= 0 */

    for(j=0; j<=M+mIntIndexMax; j++)   /* R */
        {
        /* determine QHAT */
        rbPr = (unsigned long long int)
               (RD->mIdgt[mIntIndexMin-1+j]*mIntBase + RD->mIdgt[mIntIndexMin+j]);
        if (RD->mIdgt[mIntIndexMin-1+j] >= V.mIdgt[VleftPos])
            {QHAT = mIntBaseM1;}
          else
            {QHAT = (unsigned long long int)(rbPr / V.mIdgt[VleftPos]);}
        while (1)
            {
            RHAT = (unsigned long long int)(rbPr - QHAT * V.mIdgt[VleftPos]);
            LHS = (unsigned long long int)(V.mIdgt[VleftPos+1] * QHAT);
            if (RHAT>=mIntBase) {break;}   // if carry happens in the next
            RHS = (unsigned long long int)
                  ((RHAT<<mIntBaseBits) + RD->mIdgt[mIntIndexMin-1+j+2]);

#ifdef Use_icc
            LLHLRh.I=RHAT; LLHLL.I=LHS; LLHLR.I=RHS;  //dummy code
            sprintf(dmyMsg, "R^=%08x%08x L=%08x%08x R=%08x%08x\n",
                    LLHLRh.HL.H, LLHLRh.HL.L,
                    LLHLL.HL.H, LLHLL.HL.L,
                    LLHLR.HL.H, LLHLR.HL.L
                    );                                //dummy code
#endif

            if (LHS > RHS)
                {QHAT--;}
              else
                {break;}
            }

        /* MUL & SUB */
        IHLu.I = 0;
        IHLv.I = 0;
        c=0;
        for(i=mIntIndexMax; i>=VleftPos-1; i--)  /* V */
           {
           iPj = i+j-M;   //iPj = i-(VleftPos-1)+j+(mIntIndexMin-1);
           IHLu.HL.L = RD->mIdgt[iPj];
           MulSubWk.I  = IHLu.I;                                 /* u */
           MulSubWk.I  -= c;
           IHLv.HL.L = V.mIdgt[i];                               /* v */
           vq = (unsigned long long int)(IHLv.I*QHAT);           /* v*q */
           c = (unsigned long long int)(vq >> mIntBaseBits);     /* vqH */
           vqL = (unsigned long long int)(vq & mIntBaseM1);
           MulSubWk.I -= vqL;
           RD->mIdgt[iPj] = MulSubWk.HL.L;
           if ((int)MulSubWk.HL.H < 0)
             {
             MulSubWk.HL.L = MulSubWk.HL.H;
             MulSubWk.HL.H = mIntBaseM1;
             c -= MulSubWk.I;
            }
           }
       QD->mIdgt[mIntIndexMax-M+j] = (unsigned int)QHAT;

       if(c != 0)  // Ex. in radix 10  u=8640  v=433,  q^=20  --> q=19
           {                           /*  kuwae-gaeshi */
           QD->mIdgt[mIntIndexMax-M+j]--;
           MulSubWk.I = 0;
           for(i=mIntIndexMax; i>=VleftPos-1; i--)    /* V */
               {
               iPj = i+j-M;
               IHLu.HL.L = RD->mIdgt[iPj];
               MulSubWk.I  += IHLu.I;
               IHLv.HL.L = V.mIdgt[i];
               MulSubWk.I += IHLv.I;
               RD->mIdgt[iPj] = MulSubWk.HL.L;
               MulSubWk.I >>= mIntBaseBits;
               }
           }
       }                            /* end of loop j */

    /* adjust remainder */
    if (D != 0)
        {
        for (i=mIntIndexMax*2; i>=mIntIndexMin; i--)
            {
            RD->mIdgt[i] =
                 (unsigned int)((RD->mIdgt[i]>>D) | (RD->mIdgt[i-1]<<D));
            }
        RD->mIdgt[mIntIndexMin-1] >>= D;
        }
    return(0);
    }

/*================================================================*/

/*******************************************************/
unsigned int SSI32KBin32(void)
    {
    unsigned int RanBits;
    SSI32plusRS(&SIParaSSI32K);
    RanBits=SSI32K(&SIParaSSI32K);
    return(RanBits);
    }
unsigned int SSI32KscBin32(void)
    {
    unsigned int RanBits;
    SSI32plusRS(&SIParaSSI32K);
    RanBits=SSI32Ksc(&SIParaSSI32K);
    return(RanBits);
    }
double SSI32KU01(void)
    {
    uiLLHL_Dbl SSI32Dbl;
    SSI32plusRS(&SIParaSSI32K);
    SSI32Dbl.HL.H=SSI32K(&SIParaSSI32K);
    SIParaSSI32K.wkLL.I <<= 16;
    SSI32Dbl.HL.L=SIParaSSI32K.wkLL.HL.H;
    SSI32Dbl.I >>= 12;
    SSI32Dbl.I |= 0x3ff0000000000000;     /*set exponent 2^0*/
    return(SSI32Dbl.Dbl-1);
    }
double SSI32KscU01(void)
    {
    uiLLHL_Dbl SSI32Dbl;
    SSI32plusRS(&SIParaSSI32K);
    SSI32Dbl.HL.H=SSI32Ksc(&SIParaSSI32K);
    SIParaSSI32K.wkLL.I <<= 16;
    SSI32Dbl.HL.L=SIParaSSI32K.wkLL.HL.H;
    SSI32Dbl.I >>= 12;
    SSI32Dbl.I |= 0x3ff0000000000000;     /*set exponent 2^0*/
    return(SSI32Dbl.Dbl-1);
    }
unsigned int SSI32KDec8(void)
    {
    uiLLHL_Dbl SSI32Dbl;
    SSI32plusRS(&SIParaSSI32K);
    SSI32Dbl.HL.H=SSI32K(&SIParaSSI32K);
    SIParaSSI32K.wkLL.I <<= 16;
    SSI32Dbl.HL.L=SIParaSSI32K.wkLL.HL.H;
    SSI32Dbl.I >>= 12;
    SSI32Dbl.I |= 0x3ff0000000000000;     /*set exponent 2^0*/
    return((unsigned int)((SSI32Dbl.Dbl-1)*100000000));
    }
unsigned int SSI32KscDec8(void)
    {
    uiLLHL_Dbl SSI32Dbl;
    SSI32plusRS(&SIParaSSI32K);
    SSI32Dbl.HL.H=SSI32Ksc(&SIParaSSI32K);
    SIParaSSI32K.wkLL.I <<= 16;
    SSI32Dbl.HL.L=SIParaSSI32K.wkLL.HL.H;
    SSI32Dbl.I >>= 12;
    SSI32Dbl.I |= 0x3ff0000000000000;     /*set exponent 2^0*/
    return((unsigned int)((SSI32Dbl.Dbl-1)*100000000));
    }
void SSI32IniK(unsigned long long int n)
    {
    SSI32Ini(&SIParaSSI32K, n);
    }
void SSI32IniKnm(unsigned long long int n, unsigned long long int m)
    {
    SSI32IniNM(&SIParaSSI32K, n, m);
    }
void SSI32IniKnml(unsigned long long int n, unsigned long long int m,
                  unsigned long long int l)
    {
    SSI32IniNML(&SIParaSSI32K, n, m, l);
    }

#ifdef HowToUseSSI32
/*********** How to use SSI32 **************************/
int SampleMain(void)
    {
    unsigned int ranBin32, ranDec8;
    double   ranU01;
    unsigned long long int n, m, l;

//using SSI32KBin32
    SSI32IniK(0);
    ranBin32=SSI32KBin32();             //the 0th random number
    printf("[KB1] SSI32KBin32  %08x (=76d44f81)\n", ranBin32);
    ranBin32=SSI32KBin32();             //the 1st random number
    printf("[KB2] SSI32KBin32  %08x (=aac718ec)\n", ranBin32);
    ranBin32=SSI32KBin32();             //the 2nd random number
    printf("[KB3] SSI32KBin32  %08x (=9d91efc0)\n", ranBin32);

    n=0x7ffffffe1;  m=0x7ffffffcf;
    SSI32IniKnm(n, m);            //The nm-th random number is
    ranBin32=SSI32KBin32();        // the 0th random number after 1 period
    printf("[KB4] SSI32KBin32  %08x (=76d44f81)\n", ranBin32);
    ranBin32=SSI32KBin32();             //the 1st random number
    printf("[KB5] SSI32KBin32  %08x (=aac718ec)\n", ranBin32);

    n=0x7ffffffe0;  m=0x7ffffffcf;  l=0x7ffffffcf;
    SSI32IniKnml(n, m, l);        //The (nm+l)-th random number is
    ranBin32=SSI32KBin32();        // the 0th random number after 1 period
    printf("[KB6] SSI32KBin32  %08x (=76d44f81)\n", ranBin32);

    printf("\n");

//using SSI32KU01
    SSI32IniK(0);
    ranU01=SSI32KU01();                 //the 0th random number
    printf("[KU1] SSI32KU01  %11.9f (=0.464177102)\n", ranU01);
    ranU01=SSI32KU01();                 //the 1st random number
    printf("[KU2] SSI32KU01  %11.9f (=0.667100485)\n", ranU01);
    ranU01=SSI32KU01();                 //the 2nd random number
    printf("[KU3] SSI32KU01  %11.9f (=0.615508065)\n", ranU01);

    printf("\n");

//using SSI32KDec8
    SSI32IniK(0);
    ranDec8=SSI32KDec8();               //the 0th random number
    printf("[KD1] SSI32KDec8   %08d (=46417710)\n", ranDec8);
    ranDec8=SSI32KDec8();               //the 1st random number
    printf("[KD2] SSI32KDec8   %08d (=66710048)\n", ranDec8);
    ranDec8=SSI32KDec8();               //the 2nd random number
    printf("[KD3] SSI32KDec8   %08d (=61550806)\n", ranDec8);

    printf("\n");

//using SSI32KscBin32
    SSI32IniK(0);
    ranBin32=SSI32KscBin32();           //the 0th random number
    printf("[KB1sc] SSI32KscBin32  %08x (=7e0f2267)\n", ranBin32);
    ranBin32=SSI32KscBin32();           //the 1st random number
    printf("[KB2sc] SSI32KscBin32  %08x (=561f34da)\n", ranBin32);
    ranBin32=SSI32KscBin32();           //the 2nd random number
    printf("[KB3sc] SSI32KscBin32  %08x (=d0e82521)\n", ranBin32);

    return(0);
    }

int main()
    {
    SampleMain();   //the sequence # of RNs is #0

    printf("\nChange the sequence # of Random No's to 1000.\n      (=...) should be read as (!=...)\n");
    SSI32ChangeRandSeqK(1000);
    SampleMain();   //the sequence # of RNs is changed to #1000

    printf("\nReturn to the original sequence of random No's.\n");
    SSI32ChangeRandSeqK(0);
    SampleMain();   //the sequence # of RNs is returned to #0

    return(0);
    }
#endif
